@@ -1,27 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #include "TypePrettyPrinter.h"
 
@@ -138,6 +116,9 @@ void TypePrettyPrinter::visit(UndefinedType *)
             _text.prepend(QLatin1String("signed"));
         else if (_fullySpecifiedType.isUnsigned())
             _text.prepend(QLatin1String("unsigned"));
+    } else if (_fullySpecifiedType.isAuto()) {
+        prependSpaceUnlessBracket();
+        _text.prepend("auto");
     }
 
     prependCv(_fullySpecifiedType);
@@ -169,7 +150,7 @@ void TypePrettyPrinter::visit(Template *type)
         const Overview &oo = *overview();
         if (oo.showTemplateParameters && ! _name.isEmpty()) {
             _name += QLatin1Char('<');
-            for (unsigned index = 0; index < type->templateParameterCount(); ++index) {
+            for (int index = 0; index < type->templateParameterCount(); ++index) {
                 if (index)
                     _name += QLatin1String(", ");
                 QString arg = oo.prettyName(type->templateParameterAt(index)->name());
@@ -312,8 +293,8 @@ void TypePrettyPrinter::visit(PointerToMemberType *type)
 
 void TypePrettyPrinter::prependSpaceBeforeIndirection(const FullySpecifiedType &type)
 {
-    const bool elementTypeIsPointerOrReference = type.type()->isPointerType()
-        || type.type()->isReferenceType();
+    const bool elementTypeIsPointerOrReference = type.type()->asPointerType()
+        || type.type()->asReferenceType();
     const bool elementIsConstPointerOrReference = elementTypeIsPointerOrReference && type.isConst();
     const bool shouldBindToLeftSpecifier = _overview->starBindFlags & Overview::BindToLeftSpecifier;
     if (elementIsConstPointerOrReference && ! shouldBindToLeftSpecifier)
@@ -335,15 +316,15 @@ void TypePrettyPrinter::prependSpaceAfterIndirection(bool hasName)
     const bool case2 = ! hasCvSpecifier && spaceBeforeNameNeeded;
     // case 3: In "char *argv[]", put a space between '*' and "argv" when requested
     const bool case3 = ! hasCvSpecifier && ! shouldBindToIdentifier
-        && ! _isIndirectionToArrayOrFunction && _text.size() && _text.at(0).isLetter();
+        && ! _isIndirectionToArrayOrFunction && !_text.isEmpty() && _text.at(0).isLetter();
     if (case1 || case2 || case3)
         _text.prepend(QLatin1Char(' '));
 }
 
 void TypePrettyPrinter::visit(PointerType *type)
 {
-    const bool isIndirectionToFunction = type->elementType().type()->isFunctionType();
-    const bool isIndirectionToArray = type->elementType().type()->isArrayType();
+    const bool isIndirectionToFunction = type->elementType().type()->asFunctionType();
+    const bool isIndirectionToArray = type->elementType().type()->asArrayType();
 
     visitIndirectionType(aPointerType, type->elementType(),
         isIndirectionToFunction || isIndirectionToArray);
@@ -351,8 +332,8 @@ void TypePrettyPrinter::visit(PointerType *type)
 
 void TypePrettyPrinter::visit(ReferenceType *type)
 {
-    const bool isIndirectionToFunction = type->elementType().type()->isFunctionType();
-    const bool isIndirectionToArray = type->elementType().type()->isArrayType();
+    const bool isIndirectionToFunction = type->elementType().type()->asFunctionType();
+    const bool isIndirectionToArray = type->elementType().type()->asArrayType();
     const IndirectionType indirectionType = type->isRvalueReference()
         ? aRvalueReferenceType : aReferenceType;
 
@@ -383,6 +364,39 @@ static bool endsWithPtrOrRef(const QString &type)
 
 void TypePrettyPrinter::visit(Function *type)
 {
+    bool showTemplateParameters = _overview->showTemplateParameters;
+    QStringList nameParts = _name.split("::");
+    Scope *s = type->enclosingScope();
+    if (s && s->asTemplate())
+        s = s->enclosingScope();
+
+    for (int i = nameParts.length() - 1; s && i >= 0; s = s->enclosingScope()) {
+        if (s->asClass())
+            showTemplateParameters = true;
+
+        if (Template *templ = s->asTemplate(); templ && showTemplateParameters) {
+            QString &n = nameParts[i];
+            const int paramCount = templ->templateParameterCount();
+            if (paramCount > 0) {
+                n += '<';
+                for (int index = 0; index < paramCount; ++index) {
+                    if (index)
+                        n += QLatin1String(", ");
+                    QString arg = _overview->prettyName(templ->templateParameterAt(index)->name());
+                    if (arg.isEmpty()) {
+                        arg += 'T';
+                        arg += QString::number(index + 1);
+                    }
+                    n += arg;
+                }
+                n += '>';
+            }
+        } else if (s->identifier()) {
+            --i;
+        }
+    }
+    _name = nameParts.join("::");
+
     if (_needsParens) {
         _text.prepend(QLatin1Char('('));
         if (! _name.isEmpty()) {
@@ -398,46 +412,64 @@ void TypePrettyPrinter::visit(Function *type)
         _name.clear();
     }
 
+    Overview retAndArgOverview;
+    retAndArgOverview.starBindFlags = _overview->starBindFlags;
+    retAndArgOverview.showReturnTypes = true;
+    retAndArgOverview.showArgumentNames = false;
+    retAndArgOverview.showFunctionSignatures = true;
+    retAndArgOverview.showTemplateParameters = true;
+
     if (_overview->showReturnTypes) {
-        const QString returnType = _overview->prettyType(type->returnType());
-        if (!returnType.isEmpty()) {
-            if (!endsWithPtrOrRef(returnType) || !(_overview->starBindFlags & Overview::BindToIdentifier))
-                _text.prepend(QLatin1Char(' '));
-            _text.prepend(returnType);
+        if (_overview->trailingReturnType) {
+            _text.prepend("auto ");
+        } else {
+            const QString returnType = retAndArgOverview.prettyType(type->returnType());
+            if (!returnType.isEmpty()) {
+                if (!endsWithPtrOrRef(returnType)
+                        || !(_overview->starBindFlags & Overview::BindToIdentifier)) {
+                    _text.prepend(QLatin1Char(' '));
+                }
+                _text.prepend(returnType);
+            }
         }
     }
 
     if (_overview->showEnclosingTemplate) {
-        if (Template *templ = type->enclosingTemplate()) {
-            QString templateScope = "template<";
-            for (unsigned i = 0, total = templ->templateParameterCount(); i < total; ++i) {
-                if (Symbol *param = templ->templateParameterAt(i)) {
-                    if (i > 0)
-                        templateScope.append(", ");
-                    if (TypenameArgument *typenameArg = param->asTypenameArgument()) {
-                        templateScope.append(QLatin1String(typenameArg->isClassDeclarator()
-                                                           ? "class " : "typename "));
-                        templateScope.append(_overview->prettyName(typenameArg->name()));
-                    } else if (Argument *arg = param->asArgument()) {
-                        templateScope.append(operator()(arg->type(),
-                                                        _overview->prettyName(arg->name())));
+        for (auto [s, i] = std::tuple{type->enclosingScope(), nameParts.length() - 1}; s && i >= 0;
+             s = s->enclosingScope()) {
+            if (Template *templ = s->asTemplate()) {
+                QString templateScope = "template<";
+                const int paramCount = templ->templateParameterCount();
+                for (int i = 0; i < paramCount; ++i) {
+                    if (Symbol *param = templ->templateParameterAt(i)) {
+                        if (i > 0)
+                            templateScope.append(", ");
+                        if (TypenameArgument *typenameArg = param->asTypenameArgument()) {
+                            templateScope.append(QLatin1String(
+                                typenameArg->isClassDeclarator() ? "class " : "typename "));
+                            QString name = _overview->prettyName(typenameArg->name());
+                            if (name.isEmpty())
+                                name.append('T').append(QString::number(i + 1));
+                            templateScope.append(name);
+                        } else if (TemplateTypeArgument *arg = param->asTemplateTypeArgument()) {
+                            templateScope.append(_overview->prettyName(arg->conceptName()))
+                                    .append(' ').append(_overview->prettyName(arg->name()));
+                        } else if (Argument *arg = param->asArgument()) {
+                            templateScope.append(operator()(arg->type(),
+                                                            _overview->prettyName(arg->name())));
+                        }
                     }
                 }
+                if (paramCount > 0)
+                    _text.prepend(templateScope + ">\n");
             }
-            _text.prepend(templateScope + ">\n");
         }
     }
 
     if (_overview->showFunctionSignatures) {
-        Overview argumentText;
-        argumentText.starBindFlags = _overview->starBindFlags;
-        argumentText.showReturnTypes = true;
-        argumentText.showArgumentNames = false;
-        argumentText.showFunctionSignatures = true;
-
         _text += QLatin1Char('(');
 
-        for (unsigned index = 0, argc = type->argumentCount(); index < argc; ++index) {
+        for (int index = 0, argc = type->argumentCount(); index < argc; ++index) {
             if (index != 0)
                 _text += QLatin1String(", ");
 
@@ -445,12 +477,12 @@ void TypePrettyPrinter::visit(Function *type)
                 if (index + 1 == _overview->markedArgument)
                     const_cast<Overview*>(_overview)->markedArgumentBegin = _text.length();
 
-                const Name *name = 0;
+                const Name *name = nullptr;
 
                 if (_overview->showArgumentNames)
                     name = arg->name();
 
-                _text += argumentText.prettyType(arg->type(), name);
+                _text += retAndArgOverview.prettyType(arg->type(), name);
 
                 if (_overview->showDefaultArguments) {
                     if (const StringLiteral *initializer = arg->initializer()) {
@@ -487,6 +519,21 @@ void TypePrettyPrinter::visit(Function *type)
                         ? QLatin1String("&")
                         : QLatin1String("&&");
         }
+
+        // add exception specifier
+        if (const StringLiteral *spec = type->exceptionSpecification()) {
+            if (type->refQualifier() != Function::NoRefQualifier)
+                _text.append(' ');
+            else
+                appendSpace();
+            _text += QLatin1String(spec->chars());
+        }
+    }
+
+    if (_overview->showReturnTypes && _overview->trailingReturnType) {
+        const QString returnType = retAndArgOverview.prettyType(type->returnType());
+        if (!returnType.isEmpty())
+            _text.append(" -> ").append(returnType);
     }
 }
 
