@@ -1,43 +1,23 @@
-/****************************************************************************
-**
-** Copyright (C) 2019 Denis Shienkov <denis.shienkov@gmail.com>
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2019 Denis Shienkov <denis.shienkov@gmail.com>
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+
+#include "sdcctoolchain.h"
 
 #include "baremetalconstants.h"
-
+#include "baremetaltr.h"
 #include "sdccparser.h"
-#include "sdcctoolchain.h"
 
 #include <projectexplorer/abiwidget.h>
 #include <projectexplorer/projectexplorerconstants.h>
 #include <projectexplorer/projectmacro.h>
+#include <projectexplorer/toolchainconfigwidget.h>
 #include <projectexplorer/toolchainmanager.h>
 
 #include <utils/algorithm.h>
 #include <utils/environment.h>
 #include <utils/pathchooser.h>
-#include <utils/qtcassert.h>
 #include <utils/qtcprocess.h>
+#include <utils/qtcassert.h>
 
 #include <QDebug>
 #include <QDir>
@@ -53,16 +33,7 @@
 using namespace ProjectExplorer;
 using namespace Utils;
 
-namespace BareMetal {
-namespace Internal {
-
-// Helpers:
-
-static bool compilerExists(const FilePath &compilerPath)
-{
-    const QFileInfo fi = compilerPath.toFileInfo();
-    return fi.exists() && fi.isExecutable() && fi.isFile();
-}
+namespace BareMetal::Internal {
 
 static QString compilerTargetFlag(const Abi &abi)
 {
@@ -87,14 +58,13 @@ static Macros dumpPredefinedMacros(const FilePath &compiler, const Environment &
         return {};
     fakeIn.close();
 
-    QtcProcess cpp;
+    Process cpp;
     cpp.setEnvironment(env);
-    cpp.setTimeoutS(10);
     cpp.setCommand({compiler, {compilerTargetFlag(abi),  "-dM", "-E", fakeIn.fileName()}});
 
     cpp.runBlocking();
-    if (cpp.result() != QtcProcess::FinishedWithSuccess) {
-        qWarning() << cpp.exitMessage();
+    if (cpp.result() != ProcessResult::FinishedWithSuccess) {
+        qWarning() << cpp.exitMessage(Process::FailureMessageFormat::WithStdErr);
         return {};
     }
 
@@ -108,14 +78,13 @@ static HeaderPaths dumpHeaderPaths(const FilePath &compiler, const Environment &
     if (!compiler.exists())
         return {};
 
-    QtcProcess cpp;
+    Process cpp;
     cpp.setEnvironment(env);
-    cpp.setTimeoutS(10);
     cpp.setCommand({compiler, {compilerTargetFlag(abi), "--print-search-dirs"}});
 
     cpp.runBlocking();
-    if (cpp.result() != QtcProcess::FinishedWithSuccess) {
-        qWarning() << cpp.exitMessage();
+    if (cpp.result() != ProcessResult::FinishedWithSuccess) {
+        qWarning() << cpp.exitMessage(Process::FailureMessageFormat::WithStdErr);
         return {};
     }
 
@@ -135,7 +104,7 @@ static HeaderPaths dumpHeaderPaths(const FilePath &compiler, const Environment &
             } else {
                 const QString headerPath = QFileInfo(line.trimmed())
                         .canonicalFilePath();
-                headerPaths.append({headerPath, HeaderPathType::BuiltIn});
+                headerPaths.append(HeaderPath::makeBuiltIn(headerPath));
             }
         }
     }
@@ -185,37 +154,77 @@ static Abi::BinaryFormat guessFormat(Abi::Architecture arch)
 
 static Abi guessAbi(const Macros &macros)
 {
-    const auto arch = guessArchitecture(macros);
+    const Abi::Architecture arch = guessArchitecture(macros);
     return {arch, Abi::OS::BareMetalOS, Abi::OSFlavor::GenericFlavor,
             guessFormat(arch), guessWordWidth(macros)};
 }
 
-static QString buildDisplayName(Abi::Architecture arch, Utils::Id language,
-                                const QString &version)
+static QString buildDisplayName(Abi::Architecture arch, Id language, const QString &version)
 {
-    const auto archName = Abi::toString(arch);
-    const auto langName = ToolChainManager::displayNameOfLanguageId(language);
-    return SdccToolChain::tr("SDCC %1 (%2, %3)")
-            .arg(version, langName, archName);
+    const QString archName = Abi::toString(arch);
+    const QString langName = ToolchainManager::displayNameOfLanguageId(language);
+    return Tr::tr("SDCC %1 (%2, %3)").arg(version, langName, archName);
 }
 
-static Utils::FilePath compilerPathFromEnvironment(const QString &compilerName)
+static FilePath compilerPathFromEnvironment(const QString &compilerName)
 {
     const Environment systemEnvironment = Environment::systemEnvironment();
     return systemEnvironment.searchInPath(compilerName);
 }
 
-// SdccToolChain
+// SdccToolchainConfigWidget
 
-SdccToolChain::SdccToolChain() :
-    ToolChain(Constants::SDCC_TOOLCHAIN_TYPEID)
+class SdccToolchain;
+
+class SdccToolchainConfigWidget final : public ToolchainConfigWidget
 {
-    setTypeDisplayName(Internal::SdccToolChain::tr("SDCC"));
-    setTargetAbiKey("TargetAbi");
-    setCompilerCommandKey("CompilerPath");
-}
+public:
+    explicit SdccToolchainConfigWidget(const ToolchainBundle &bundle);
 
-ToolChain::MacroInspectionRunner SdccToolChain::createMacroInspectionRunner() const
+private:
+    void applyImpl() final;
+    void discardImpl() final { setFromToolchain(); }
+    bool isDirtyImpl() const final;
+    void makeReadOnlyImpl() final;
+
+    void setFromToolchain();
+    void handleCompilerCommandChange();
+
+    AbiWidget *m_abiWidget = nullptr;
+    Macros m_macros;
+};
+
+// SdccToolchain
+
+class SdccToolchain final : public Toolchain
+{
+public:
+    SdccToolchain() : Toolchain(Constants::SDCC_TOOLCHAIN_TYPEID)
+    {
+        setTypeDisplayName(Tr::tr("SDCC"));
+        setTargetAbiKey("TargetAbi");
+        setCompilerCommandKey("CompilerPath");
+    }
+
+    MacroInspectionRunner createMacroInspectionRunner() const final;
+
+    LanguageExtensions languageExtensions(const QStringList &cxxflags) const final;
+    WarningFlags warningFlags(const QStringList &cxxflags) const final;
+
+    BuiltInHeaderPathsRunner createBuiltInHeaderPathsRunner(const Environment &) const final;
+    void addToEnvironment(Environment &env) const final;
+    QList<OutputLineParser *> createOutputParsers() const final  { return {new SdccParser}; }
+
+    bool operator==(const Toolchain &other) const final;
+
+    FilePath makeCommand(const Environment &) const final { return {}; }
+
+private:
+    friend class SdccToolchainFactory;
+    friend class SdccToolchainConfigWidget;
+};
+
+Toolchain::MacroInspectionRunner SdccToolchain::createMacroInspectionRunner() const
 {
     Environment env = Environment::systemEnvironment();
     addToEnvironment(env);
@@ -238,18 +247,18 @@ ToolChain::MacroInspectionRunner SdccToolChain::createMacroInspectionRunner() co
     };
 }
 
-Utils::LanguageExtensions SdccToolChain::languageExtensions(const QStringList &) const
+LanguageExtensions SdccToolchain::languageExtensions(const QStringList &) const
 {
     return LanguageExtension::None;
 }
 
-WarningFlags SdccToolChain::warningFlags(const QStringList &cxxflags) const
+WarningFlags SdccToolchain::warningFlags(const QStringList &cxxflags) const
 {
     Q_UNUSED(cxxflags)
     return WarningFlags::Default;
 }
 
-ToolChain::BuiltInHeaderPathsRunner SdccToolChain::createBuiltInHeaderPathsRunner(
+Toolchain::BuiltInHeaderPathsRunner SdccToolchain::createBuiltInHeaderPathsRunner(
         const Environment &) const
 {
     Environment env = Environment::systemEnvironment();
@@ -258,60 +267,61 @@ ToolChain::BuiltInHeaderPathsRunner SdccToolChain::createBuiltInHeaderPathsRunne
     const FilePath compiler = compilerCommand();
     const Abi abi = targetAbi();
 
-    return [env, compiler, abi](const QStringList &, const QString &, const QString &) {
+    return [env, compiler, abi](const QStringList &, const FilePath &, const QString &) {
         return dumpHeaderPaths(compiler, env, abi);
     };
 }
 
-void SdccToolChain::addToEnvironment(Environment &env) const
+void SdccToolchain::addToEnvironment(Environment &env) const
 {
-    if (!compilerCommand().isEmpty()) {
-        const FilePath path = compilerCommand().parentDir();
-        env.prependOrSetPath(path.toString());
-    }
+    if (!compilerCommand().isEmpty())
+        env.prependOrSetPath(compilerCommand().parentDir());
 }
 
-QList<Utils::OutputLineParser *> SdccToolChain::createOutputParsers() const
+bool SdccToolchain::operator==(const Toolchain &other) const
 {
-    return {new SdccParser};
-}
-
-std::unique_ptr<ToolChainConfigWidget> SdccToolChain::createConfigurationWidget()
-{
-    return std::make_unique<SdccToolChainConfigWidget>(this);
-}
-
-bool SdccToolChain::operator==(const ToolChain &other) const
-{
-    if (!ToolChain::operator==(other))
+    if (!Toolchain::operator==(other))
         return false;
 
-    const auto customTc = static_cast<const SdccToolChain *>(&other);
+    const auto customTc = static_cast<const SdccToolchain *>(&other);
     return compilerCommand() == customTc->compilerCommand()
             && targetAbi() == customTc->targetAbi();
 }
 
-FilePath SdccToolChain::makeCommand(const Environment &env) const
+// SdccToolchainFactory
+
+class SdccToolchainFactory final : public ToolchainFactory
 {
-    Q_UNUSED(env)
-    return {};
+public:
+    SdccToolchainFactory()
+    {
+        setDisplayName(Tr::tr("SDCC"));
+        setSupportedToolchainType(Constants::SDCC_TOOLCHAIN_TYPEID);
+        setSupportedLanguages({ProjectExplorer::Constants::C_LANGUAGE_ID});
+        setToolchainConstructor([] { return new SdccToolchain; });
+        setUserCreatable(true);
+    }
+
+    Toolchains autoDetect(const ToolchainDetector &detector) const final;
+    std::unique_ptr<ToolchainConfigWidget> createConfigurationWidget(
+        const ToolchainBundle &bundle) const final
+    {
+        return std::make_unique<SdccToolchainConfigWidget>(bundle);
+    }
+
+private:
+    Toolchains autoDetectToolchains(const Candidates &candidates,
+                                    const Toolchains &alreadyKnown) const;
+    Toolchains autoDetectToolchain(const Candidate &candidate, Id language) const;
+};
+
+void setupSdccToolchain()
+{
+    static SdccToolchainFactory theSdccToolchainFactory;
 }
 
-// SdccToolChainFactory
-
-SdccToolChainFactory::SdccToolChainFactory()
+Toolchains SdccToolchainFactory::autoDetect(const ToolchainDetector &detector) const
 {
-    setDisplayName(SdccToolChain::tr("SDCC"));
-    setSupportedToolChainType(Constants::SDCC_TOOLCHAIN_TYPEID);
-    setSupportedLanguages({ProjectExplorer::Constants::C_LANGUAGE_ID});
-    setToolchainConstructor([] { return new SdccToolChain; });
-    setUserCreatable(true);
-}
-
-QList<ToolChain *> SdccToolChainFactory::autoDetect(const QList<ToolChain *> &alreadyKnown,
-                                                    const IDevice::Ptr &device)
-{
-    Q_UNUSED(device)
     Candidates candidates;
 
     if (Utils::HostOsInfo::isWindowsHost()) {
@@ -328,7 +338,7 @@ QList<ToolChain *> SdccToolChainFactory::autoDetect(const QList<ToolChain *> &al
             compilerPath += "/bin/sdcc.exe";
             const FilePath fn = FilePath::fromString(
                         QFileInfo(compilerPath).absoluteFilePath());
-            if (!compilerExists(fn))
+            if (!fn.isExecutableFile())
                 return Candidate{};
             // Build compiler version.
             const QString version = QString("%1.%2.%3").arg(
@@ -364,17 +374,16 @@ QList<ToolChain *> SdccToolChainFactory::autoDetect(const QList<ToolChain *> &al
             candidates.push_back(candidate);
     }
 
-    return autoDetectToolchains(candidates, alreadyKnown);
+    return autoDetectToolchains(candidates, detector.alreadyKnown);
 }
 
-QList<ToolChain *> SdccToolChainFactory::autoDetectToolchains(
-        const Candidates &candidates, const QList<ToolChain *> &alreadyKnown) const
+Toolchains SdccToolchainFactory::autoDetectToolchains(
+        const Candidates &candidates, const Toolchains &alreadyKnown) const
 {
-    QList<ToolChain *> result;
+    Toolchains result;
 
-    for (const Candidate &candidate : qAsConst(candidates)) {
-        const QList<ToolChain *> filtered = Utils::filtered(
-                    alreadyKnown, [candidate](ToolChain *tc) {
+    for (const Candidate &candidate : std::as_const(candidates)) {
+        const Toolchains filtered = Utils::filtered(alreadyKnown, [candidate](Toolchain *tc) {
             return tc->typeId() == Constants::SDCC_TOOLCHAIN_TYPEID
                 && tc->compilerCommand() == candidate.compilerPath
                 && (tc->language() == ProjectExplorer::Constants::C_LANGUAGE_ID);
@@ -392,8 +401,7 @@ QList<ToolChain *> SdccToolChainFactory::autoDetectToolchains(
     return result;
 }
 
-QList<ToolChain *> SdccToolChainFactory::autoDetectToolchain(
-        const Candidate &candidate, Utils::Id language) const
+Toolchains SdccToolchainFactory::autoDetectToolchain(const Candidate &candidate, Id language) const
 {
     const auto env = Environment::systemEnvironment();
 
@@ -403,7 +411,7 @@ QList<ToolChain *> SdccToolChainFactory::autoDetectToolchain(
         {Abi::Stm8Architecture}
     };
 
-    QList<ToolChain *> tcs;
+    Toolchains tcs;
 
     // Probe each ABI from the table, because the SDCC compiler
     // can be compiled with or without the specified architecture.
@@ -415,15 +423,15 @@ QList<ToolChain *> SdccToolChainFactory::autoDetectToolchain(
         if (knownAbi.architecture() != abi.architecture())
             continue;
 
-        const auto tc = new SdccToolChain;
-        tc->setDetection(ToolChain::AutoDetection);
+        const auto tc = new SdccToolchain;
+        tc->setDetection(Toolchain::AutoDetection);
         tc->setLanguage(language);
         tc->setCompilerCommand(candidate.compilerPath);
         tc->setTargetAbi(abi);
         tc->setDisplayName(buildDisplayName(abi.architecture(), language,
                                             candidate.compilerVersion));
 
-        const auto languageVersion = ToolChain::languageVersion(language, macros);
+        const auto languageVersion = Toolchain::languageVersion(language, macros);
         tc->predefinedMacrosCache()->insert({}, {macros, languageVersion});
 
         tcs.push_back(tc);
@@ -432,77 +440,64 @@ QList<ToolChain *> SdccToolChainFactory::autoDetectToolchain(
     return tcs;
 }
 
-// SdccToolChainConfigWidget
+// SdccToolchainConfigWidget
 
-SdccToolChainConfigWidget::SdccToolChainConfigWidget(SdccToolChain *tc) :
-    ToolChainConfigWidget(tc),
-    m_compilerCommand(new PathChooser),
+SdccToolchainConfigWidget::SdccToolchainConfigWidget(const ToolchainBundle &bundle) :
+    ToolchainConfigWidget(bundle),
     m_abiWidget(new AbiWidget)
 {
-    m_compilerCommand->setExpectedKind(PathChooser::ExistingCommand);
-    m_compilerCommand->setHistoryCompleter("PE.SDCC.Command.History");
-    m_mainLayout->addRow(tr("&Compiler path:"), m_compilerCommand);
-    m_mainLayout->addRow(tr("&ABI:"), m_abiWidget);
+    m_mainLayout->addRow(Tr::tr("&ABI:"), m_abiWidget);
 
     m_abiWidget->setEnabled(false);
 
     addErrorLabel();
     setFromToolchain();
 
-    connect(m_compilerCommand, &PathChooser::rawPathChanged,
-            this, &SdccToolChainConfigWidget::handleCompilerCommandChange);
+    connect(this, &ToolchainConfigWidget::compilerCommandChanged,
+            this, &SdccToolchainConfigWidget::handleCompilerCommandChange);
     connect(m_abiWidget, &AbiWidget::abiChanged,
-            this, &ToolChainConfigWidget::dirty);
+            this, &ToolchainConfigWidget::dirty);
 }
 
-void SdccToolChainConfigWidget::applyImpl()
+void SdccToolchainConfigWidget::applyImpl()
 {
-    if (toolChain()->isAutoDetected())
+    if (bundle().isAutoDetected())
         return;
 
-    const auto tc = static_cast<SdccToolChain *>(toolChain());
-    const QString displayName = tc->displayName();
-    tc->setCompilerCommand(m_compilerCommand->filePath());
-    tc->setTargetAbi(m_abiWidget->currentAbi());
-    tc->setDisplayName(displayName);
-
+    bundle().setTargetAbi(m_abiWidget->currentAbi());
     if (m_macros.isEmpty())
         return;
 
-    const auto languageVersion = ToolChain::languageVersion(tc->language(), m_macros);
-    tc->predefinedMacrosCache()->insert({}, {m_macros, languageVersion});
-
+    bundle().forEach<SdccToolchain>([this](SdccToolchain &tc) {
+        const auto languageVersion = Toolchain::languageVersion(tc.language(), m_macros);
+        tc.predefinedMacrosCache()->insert({}, {m_macros, languageVersion});
+    });
     setFromToolchain();
 }
 
-bool SdccToolChainConfigWidget::isDirtyImpl() const
+bool SdccToolchainConfigWidget::isDirtyImpl() const
 {
-    const auto tc = static_cast<SdccToolChain *>(toolChain());
-    return m_compilerCommand->filePath() != tc->compilerCommand()
-            || m_abiWidget->currentAbi() != tc->targetAbi()
-            ;
+    return m_abiWidget->currentAbi() != bundle().targetAbi();
 }
 
-void SdccToolChainConfigWidget::makeReadOnlyImpl()
+void SdccToolchainConfigWidget::makeReadOnlyImpl()
 {
-    m_compilerCommand->setReadOnly(true);
     m_abiWidget->setEnabled(false);
 }
 
-void SdccToolChainConfigWidget::setFromToolchain()
+void SdccToolchainConfigWidget::setFromToolchain()
 {
     const QSignalBlocker blocker(this);
-    const auto tc = static_cast<SdccToolChain *>(toolChain());
-    m_compilerCommand->setFilePath(tc->compilerCommand());
-    m_abiWidget->setAbis({}, tc->targetAbi());
-    const bool haveCompiler = compilerExists(m_compilerCommand->filePath());
-    m_abiWidget->setEnabled(haveCompiler && !tc->isAutoDetected());
+    m_abiWidget->setAbis({}, bundle().targetAbi());
+    const bool haveCompiler
+        = compilerCommand(ProjectExplorer::Constants::C_LANGUAGE_ID).isExecutableFile();
+    m_abiWidget->setEnabled(haveCompiler && !bundle().isAutoDetected());
 }
 
-void SdccToolChainConfigWidget::handleCompilerCommandChange()
+void SdccToolchainConfigWidget::handleCompilerCommandChange()
 {
-    const FilePath compilerPath = m_compilerCommand->filePath();
-    const bool haveCompiler = compilerExists(compilerPath);
+    const FilePath compilerPath = compilerCommand(ProjectExplorer::Constants::C_LANGUAGE_ID);
+    const bool haveCompiler = compilerPath.isExecutableFile();
     if (haveCompiler) {
         const auto env = Environment::systemEnvironment();
         m_macros = dumpPredefinedMacros(compilerPath, env, {});
@@ -514,5 +509,4 @@ void SdccToolChainConfigWidget::handleCompilerCommandChange()
     emit dirty();
 }
 
-} // namespace Internal
-} // namespace BareMetal
+} // BareMetal::Internal
