@@ -1,67 +1,341 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+
+#include "progressmanager_p.h"
 
 #include "futureprogress.h"
-#include "progressmanager_p.h"
 #include "progressbar.h"
 #include "progressview.h"
 #include "../actionmanager/actionmanager.h"
 #include "../actionmanager/command.h"
+#include "../coreplugintr.h"
 #include "../icontext.h"
-#include "../coreconstants.h"
 #include "../icore.h"
 #include "../statusbarmanager.h"
 
 #include <extensionsystem/pluginmanager.h>
+
+#include <utils/environment.h>
 #include <utils/hostosinfo.h>
+#include <utils/infobar.h>
+#include <utils/layoutbuilder.h>
+#include <utils/mathutils.h>
 #include <utils/qtcassert.h>
 #include <utils/stylehelper.h>
 #include <utils/theme/theme.h>
 #include <utils/utilsicons.h>
+#include <utils/utilstr.h>
 
 #include <QAction>
+#include <QComboBox>
 #include <QEvent>
-#include <QHBoxLayout>
+#include <QFuture>
+#include <QFutureInterfaceBase>
 #include <QMouseEvent>
 #include <QPainter>
 #include <QPropertyAnimation>
+#include <QScrollArea>
 #include <QStyle>
 #include <QStyleOption>
 #include <QTimer>
 #include <QVariant>
 
-#include <math.h>
+using namespace Core::Internal;
+using namespace Utils;
+
+using namespace std::chrono;
 
 static const char kSettingsGroup[] = "Progress";
 static const char kDetailsPinned[] = "DetailsPinned";
 static const bool kDetailsPinnedDefault = true;
-static const int TimerInterval = 100; // 100 ms
+static const milliseconds TimerInterval{100};
 
-using namespace Core;
-using namespace Core::Internal;
-using namespace Utils;
+namespace Core {
+
+namespace Internal {
+
+int ProgressManagerPrivate::infoMinWidth()
+{
+    return 100;
+}
+
+int ProgressManagerPrivate::infoMaxWidth()
+{
+    return 350;
+}
+
+class PopupInfoBarDisplay : public QWidget
+{
+public:
+    PopupInfoBarDisplay();
+
+    void setInfoBar(InfoBar *infoBar);
+    InfoBar *infoBar() const;
+    void paintEvent(QPaintEvent *ev) override;
+
+private:
+    void update();
+
+    QVBoxLayout *m_layout = nullptr;
+    QPointer<InfoBar> m_infoBar;
+    QList<QWidget *> m_infoWidgets;
+};
+
+class InfoWidget : public QWidget
+{
+public:
+    InfoWidget(const InfoBarEntry &info, QPointer<InfoBar> infoBar);
+
+    void paintEvent(QPaintEvent *ev) override;
+
+private:
+    QPointer<QWidget> m_detailsWidget;
+};
+
+PopupInfoBarDisplay::PopupInfoBarDisplay()
+{
+    m_layout = new QVBoxLayout;
+    m_layout->setContentsMargins(0, 2, 0, 0);
+    m_layout->setSpacing(0);
+    setLayout(m_layout);
+}
+
+void PopupInfoBarDisplay::setInfoBar(InfoBar *infoBar)
+{
+    if (m_infoBar == infoBar)
+        return;
+
+    if (m_infoBar)
+        m_infoBar->disconnect(this);
+    m_infoBar = infoBar;
+    if (m_infoBar)
+        connect(m_infoBar.data(), &InfoBar::changed, this, &PopupInfoBarDisplay::update);
+    update();
+}
+
+InfoBar *PopupInfoBarDisplay::infoBar() const
+{
+    return m_infoBar;
+}
+
+void PopupInfoBarDisplay::paintEvent(QPaintEvent *)
+{
+    QPainter p(this);
+    const QRect r = rect().adjusted(0, layout()->contentsMargins().top(), 0, 0);
+    StyleHelper::drawCardBg(&p, r, creatorColor(Theme::Token_Background_Muted),
+                            creatorColor(Theme::Token_Stroke_Subtle),
+                            StyleHelper::SpacingTokens::HPaddingXs);
+}
+
+static void disconnectRecursively(QObject *obj)
+{
+    obj->disconnect();
+    for (QObject *child : obj->children())
+        disconnectRecursively(child);
+}
+
+void PopupInfoBarDisplay::update()
+{
+    for (QWidget *widget : std::as_const(m_infoWidgets)) {
+        // Make sure that we are no longer connect to anything (especially lambdas).
+        // Otherwise a lambda might live longer than the owner of the lambda.
+        disconnectRecursively(widget);
+
+        widget->hide(); // Late deletion can cause duplicate infos. Hide immediately to prevent it.
+        widget->deleteLater();
+    }
+    m_infoWidgets.clear();
+
+    if (!m_infoBar)
+        return;
+
+    const QList<InfoBarEntry> entries = m_infoBar->entries();
+    for (const InfoBarEntry &info : entries) {
+        auto widget = new InfoWidget(info, m_infoBar);
+        m_layout->addWidget(widget);
+        m_infoWidgets.append(widget);
+    }
+    setVisible(!entries.empty());
+}
+
+InfoWidget::InfoWidget(const InfoBarEntry &info, QPointer<InfoBar> infoBar)
+{
+    using namespace Layouting;
+
+    setMinimumWidth(ProgressManagerPrivate::infoMinWidth());
+    setMaximumWidth(ProgressManagerPrivate::infoMaxWidth());
+
+    const InfoLabel::InfoType infoType = [&info] {
+        const QString envString = qtcEnvironmentVariable("QTC_DEBUG_POPUPNOTIFICATION_TYPE", {});
+        if (!envString.isEmpty()) {
+            bool ok = false;
+            const int i = envString.toUInt(&ok);
+            if (ok && i <= int(InfoLabel::None))
+                return InfoLabel::InfoType(i);
+        }
+        return info.infoType();
+    }();
+
+    const Id id = info.id();
+
+    QToolButton *infoWidgetCloseButton = nullptr;
+    QLayout *buttonLayout;
+    QLabel *titleLabel;
+
+    // clang-format off
+    Column {
+        Row {
+            Column {
+                Layouting::IconDisplay { icon(InfoBarEntry::icon(infoType)) },
+                st,
+                customMargins(0, 0, StyleHelper::SpacingTokens::ExPaddingGapS, 0),
+            },
+            Column {
+                Row {
+                    Label { bindTo(&titleLabel), text(info.title()) },
+                    st,
+                    If { info.hasCancelButton(), { ToolButton { bindTo(&infoWidgetCloseButton) } } }
+                },
+                Row {
+                    Label { wordWrap(true), text(info.text()) }
+                }
+            }
+        },
+        Space(StyleHelper::SpacingTokens::HGapS),
+        Flow{ bindTo(&buttonLayout), alignment(Qt::AlignRight) },
+        customMargins(StyleHelper::SpacingTokens::HPaddingS,
+                      StyleHelper::SpacingTokens::ExPaddingGapL,
+                      StyleHelper::SpacingTokens::HPaddingS,
+                      StyleHelper::SpacingTokens::ExPaddingGapL),
+    }.attachTo(this);
+    // clang-format on
+
+    titleLabel->setFont(StyleHelper::uiFont(StyleHelper::UiElementH5));
+
+    if (info.hasCancelButton() && QTC_GUARD(infoWidgetCloseButton)) {
+        // need to connect to cancelObjectbefore connecting to cancelButtonClicked,
+        // because the latter removes the button and with it any connect
+        if (info.cancelButtonCallback())
+            connect(infoWidgetCloseButton, &QAbstractButton::clicked, info.cancelButtonCallback());
+        connect(infoWidgetCloseButton, &QAbstractButton::clicked, this, [infoBar, id] {
+            if (infoBar)
+                infoBar->removeInfo(id);
+        });
+        if (info.cancelButtonText().isEmpty()) {
+            infoWidgetCloseButton->setAutoRaise(true);
+            infoWidgetCloseButton->setIcon(Icons::CLOSE_FOREGROUND.icon());
+            infoWidgetCloseButton->setToolTip(Tr::tr("Close"));
+        } else {
+            infoWidgetCloseButton->setText(info.cancelButtonText());
+        }
+    }
+
+    const InfoBarEntry::Combo combo = info.combo();
+    if (!combo.entries.isEmpty()) {
+        auto cb = new QComboBox();
+        cb->setToolTip(combo.tooltip);
+        for (const InfoBarEntry::ComboInfo &comboInfo : std::as_const(combo.entries))
+            cb->addItem(comboInfo.displayText, comboInfo.data);
+        if (combo.currentIndex >= 0 && combo.currentIndex < cb->count())
+            cb->setCurrentIndex(combo.currentIndex);
+        connect(
+            cb,
+            &QComboBox::currentIndexChanged,
+            this,
+            [cb, combo] { combo.callback({cb->currentText(), cb->currentData()}); },
+            Qt::QueuedConnection);
+        buttonLayout->addWidget(cb);
+    }
+
+    if (info.detailsWidgetCreator()) {
+        auto showDetailsButton = new QToolButton;
+        showDetailsButton->setText(Tr::tr("Show Details..."));
+        connect(showDetailsButton, &QToolButton::clicked, this, [this, info](bool) {
+            if (m_detailsWidget) {
+                ICore::raiseWindow(m_detailsWidget);
+                return;
+            }
+            m_detailsWidget = new QWidget;
+            m_detailsWidget->setAttribute(Qt::WA_DeleteOnClose);
+            m_detailsWidget->setLayout(new QVBoxLayout);
+            auto scrollArea = new QScrollArea;
+            scrollArea->setWidgetResizable(true);
+            scrollArea->setWidget(info.detailsWidgetCreator()());
+            m_detailsWidget->layout()->addWidget(scrollArea);
+            m_detailsWidget->setWindowTitle(scrollArea->widget()->windowTitle());
+            ICore::registerWindow(
+                m_detailsWidget, Context(info.id().withPrefix("PopupNotification.Details.")));
+            m_detailsWidget->show();
+        });
+        buttonLayout->addWidget(showDetailsButton);
+    }
+
+    const QList<InfoBarEntry::Button> buttons = info.buttons();
+    for (const InfoBarEntry::Button &button : buttons) {
+        auto infoWidgetButton = new QToolButton;
+        infoWidgetButton->setText(button.text);
+        infoWidgetButton->setToolTip(button.tooltip);
+        connect(infoWidgetButton, &QAbstractButton::clicked, [button, infoBar, id] {
+            infoBar->triggerButton(id, button);
+        });
+        buttonLayout->addWidget(infoWidgetButton);
+    }
+
+    if (info.globalSuppression() == InfoBarEntry::GlobalSuppression::Enabled) {
+        auto infoWidgetSuppressButton = new QToolButton;
+        infoWidgetSuppressButton->setText(Utils::Tr::tr("Do Not Show Again"));
+        connect(infoWidgetSuppressButton, &QAbstractButton::clicked, this, [infoBar, id] {
+            if (!infoBar)
+                return;
+            infoBar->removeInfo(id);
+            InfoBar::globallySuppressInfo(id);
+        });
+        buttonLayout->addWidget(infoWidgetSuppressButton);
+    }
+}
+
+void InfoWidget::paintEvent(QPaintEvent *)
+{
+    QPainter p(this);
+    p.setPen(creatorColor(Theme::Token_Stroke_Subtle));
+    p.drawLine(Utils::StyleHelper::SpacingTokens::HPaddingXs, 0,
+               width() - Utils::StyleHelper::SpacingTokens::HPaddingXs - 1, 0);
+}
+
+class ProgressTimer : public QObject
+{
+public:
+    ProgressTimer(const QFutureInterfaceBase &futureInterface,
+                  seconds expectedDuration,
+                  QObject *parent)
+        : QObject(parent)
+        , m_futureInterface(futureInterface)
+        , m_expectedDuration(expectedDuration)
+    {
+        m_futureInterface.setProgressRange(0, 100);
+        m_futureInterface.setProgressValue(0);
+
+        m_timer.setInterval(TimerInterval);
+        connect(&m_timer, &QTimer::timeout, this, &ProgressTimer::handleTimeout);
+        m_timer.start();
+    }
+
+private:
+    void handleTimeout()
+    {
+        ++m_currentTime;
+        const int halfLife = m_expectedDuration / TimerInterval;
+        const int progress = MathUtils::interpolateTangential(m_currentTime, halfLife, 0, 100);
+        m_futureInterface.setProgressValue(progress);
+    }
+
+    QFutureInterfaceBase m_futureInterface;
+    seconds m_expectedDuration;
+    int m_currentTime = 0;
+    QTimer m_timer;
+};
+
+} // namespace Internal
 
 /*!
     \class Core::ProgressManager
@@ -130,7 +404,7 @@ using namespace Utils;
     start a task concurrently in a different thread.
     QtConcurrent has several different functions to run e.g.
     a class function in a different thread. Qt Creator itself
-    adds a few more in \c{src/libs/qtconcurrent/runextensions.h}.
+    adds a few more in \c{src/libs/utils/async.h}.
     The QtConcurrent functions to run a concurrent task return a
     \c QFuture object. This is what you want to give the
     ProgressManager in the addTask() function.
@@ -155,7 +429,7 @@ using namespace Utils;
     in a different thread, looks like this:
     \code
     QFuture<void> task = Utils::map(filters, &ILocatorFilter::refresh);
-    Core::FutureProgress *progress = Core::ProgressManager::addTask(task, tr("Indexing"),
+    Core::FutureProgress *progress = Core::ProgressManager::addTask(task, ::Core::Tr::tr("Indexing"),
                                                                     Locator::Constants::TASK_INDEX);
     \endcode
     First, we to start an asynchronous operation which calls all the filters'
@@ -171,7 +445,7 @@ using namespace Utils;
     // We are already running in a different thread here
     QFutureInterface<void> *progressObject = new QFutureInterface<void>;
     progressObject->setProgressRange(0, MAX);
-    Core::ProgressManager::addTask(progressObject->future(), tr("DoIt"), MYTASKTYPE);
+    Core::ProgressManager::addTask(progressObject->future(), ::Core::Tr::tr("DoIt"), MYTASKTYPE);
     progressObject->reportStarted();
     // Do something
     ...
@@ -209,7 +483,7 @@ using namespace Utils;
     \value ShowInApplicationIcon
         The progress indicator for this task is additionally
         shown in the application icon in the system's task bar or dock, on
-        platforms that support that (at the moment Windows 7 and Mac OS X).
+        platforms that support that (at the moment Windows 7 and macOS).
 */
 
 /*!
@@ -249,13 +523,21 @@ public:
 
 ProgressManagerPrivate::ProgressManagerPrivate()
     : m_opacityEffect(new QGraphicsOpacityEffect(this))
+    , m_appLabelUpdateTimer(new QTimer(this))
 {
     m_opacityEffect->setOpacity(.999);
     m_instance = this;
+
     m_progressView = new ProgressView;
+    m_infoBarDisplay = new PopupInfoBarDisplay;
+    m_infoBarDisplay->setInfoBar(ICore::popupInfoBar());
+    m_progressView->addProgressWidget(m_infoBarDisplay);
+
     // withDelay, so the statusBarWidget has the chance to get the enter event
     connect(m_progressView.data(), &ProgressView::hoveredChanged, this, &ProgressManagerPrivate::updateVisibilityWithDelay);
     connect(ICore::instance(), &ICore::coreAboutToClose, this, &ProgressManagerPrivate::cancelAllRunningTasks);
+    m_appLabelUpdateTimer->setSingleShot(true);
+    m_appLabelUpdateTimer->callOnTimeout(this, &ProgressManagerPrivate::updateApplicationLabelNow);
 }
 
 ProgressManagerPrivate::~ProgressManagerPrivate()
@@ -271,7 +553,7 @@ ProgressManagerPrivate::~ProgressManagerPrivate()
 
 void ProgressManagerPrivate::readSettings()
 {
-    QSettings *settings = ICore::settings();
+    QtcSettings *settings = ICore::settings();
     settings->beginGroup(kSettingsGroup);
     m_progressViewPinned = settings->value(kDetailsPinned, kDetailsPinnedDefault).toBool();
     settings->endGroup();
@@ -308,12 +590,21 @@ void ProgressManagerPrivate::init()
     m_summaryProgressBar->setCancelEnabled(false);
     summaryProgressLayout->addWidget(m_summaryProgressBar);
     layout->addWidget(m_summaryProgressWidget);
+    m_notificationSummaryIcon = new QLabel;
+    m_notificationSummaryIcon->setPixmap(Icons::WARNING_TOOLBAR.pixmap());
+    m_notificationSummaryIcon->setVisible(false);
+    connect(
+        m_infoBarDisplay->infoBar(),
+        &InfoBar::changed,
+        this,
+        &ProgressManagerPrivate::updateNotificationSummaryIcon);
+    layout->addWidget(m_notificationSummaryIcon);
     auto toggleButton = new QToolButton(m_statusBarWidget);
     layout->addWidget(toggleButton);
     m_statusBarWidget->installEventFilter(this);
     StatusBarManager::addStatusBarWidget(m_statusBarWidget, StatusBarManager::RightCorner);
 
-    QAction *toggleProgressView = new QAction(tr("Toggle Progress Details"), this);
+    QAction *toggleProgressView = new QAction(::Core::Tr::tr("Toggle Progress Details"), this);
     toggleProgressView->setCheckable(true);
     toggleProgressView->setChecked(m_progressViewPinned);
     toggleProgressView->setIcon(Utils::Icons::TOGGLE_PROGRESSDETAILS_TOOLBAR.icon());
@@ -333,19 +624,19 @@ void ProgressManagerPrivate::init()
 void ProgressManagerPrivate::doCancelTasks(Id type)
 {
     bool found = false;
-    QMap<QFutureWatcher<void> *, Id>::iterator task = m_runningTasks.begin();
-    while (task != m_runningTasks.end()) {
-        if (task.value() != type) {
-            ++task;
+    auto it = m_runningTasks.cbegin();
+    while (it != m_runningTasks.cend()) {
+        if (it.value() != type) {
+            ++it;
             continue;
         }
         found = true;
-        disconnect(task.key(), &QFutureWatcherBase::finished, this, &ProgressManagerPrivate::taskFinished);
-        if (m_applicationTask == task.key())
+        if (m_applicationTask == it.key())
             disconnectApplicationTask();
-        task.key()->cancel();
-        delete task.key();
-        task = m_runningTasks.erase(task);
+        it.key()->disconnect();
+        it.key()->cancel();
+        delete it.key();
+        it = m_runningTasks.erase(it);
     }
     if (found) {
         updateSummaryProgressBar();
@@ -381,14 +672,12 @@ bool ProgressManagerPrivate::eventFilter(QObject *obj, QEvent *event)
 
 void ProgressManagerPrivate::cancelAllRunningTasks()
 {
-    QMap<QFutureWatcher<void> *, Id>::const_iterator task = m_runningTasks.constBegin();
-    while (task != m_runningTasks.constEnd()) {
-        disconnect(task.key(), &QFutureWatcherBase::finished, this, &ProgressManagerPrivate::taskFinished);
-        if (m_applicationTask == task.key())
+    for (auto it = m_runningTasks.cbegin(); it != m_runningTasks.cend(); ++it) {
+        if (m_applicationTask == it.key())
             disconnectApplicationTask();
-        task.key()->cancel();
-        delete task.key();
-        ++task;
+        it.key()->disconnect();
+        it.key()->cancel();
+        delete it.key();
     }
     m_runningTasks.clear();
     updateSummaryProgressBar();
@@ -404,12 +693,13 @@ FutureProgress *ProgressManagerPrivate::doAddTask(const QFuture<void> &future, c
             this, &ProgressManagerPrivate::updateSummaryProgressBar);
     connect(watcher, &QFutureWatcherBase::progressValueChanged,
             this, &ProgressManagerPrivate::updateSummaryProgressBar);
-    connect(watcher, &QFutureWatcherBase::finished, this, &ProgressManagerPrivate::taskFinished);
+    connect(watcher, &QFutureWatcherBase::finished, this, [this, watcher] {
+        taskFinished(watcher);
+    });
 
     // handle application task
     if (flags & ShowInApplicationIcon) {
-        if (m_applicationTask)
-            disconnectApplicationTask();
+        disconnectApplicationTask();
         m_applicationTask = watcher;
         setApplicationProgressRange(future.progressMinimum(), future.progressMaximum());
         setApplicationProgressValue(future.progressValue());
@@ -439,7 +729,11 @@ FutureProgress *ProgressManagerPrivate::doAddTask(const QFuture<void> &future, c
         progress->setKeepOnFinish(FutureProgress::HideOnFinish);
     connect(progress, &FutureProgress::hasErrorChanged,
             this, &ProgressManagerPrivate::updateSummaryProgressBar);
-    connect(progress, &FutureProgress::removeMe, this, &ProgressManagerPrivate::slotRemoveTask);
+    connect(progress, &FutureProgress::removeMe, this, [this, progress] {
+        const Id type = progress->type();
+        removeTask(progress);
+        removeOldTasks(type, true);
+    });
     connect(progress, &FutureProgress::fadeStarted,
             this, &ProgressManagerPrivate::updateSummaryProgressBar);
     connect(progress, &FutureProgress::statusBarWidgetChanged,
@@ -457,24 +751,27 @@ ProgressView *ProgressManagerPrivate::progressView()
     return m_progressView;
 }
 
-void ProgressManagerPrivate::taskFinished()
+void ProgressManagerPrivate::taskFinished(QFutureWatcher<void> *task)
 {
-    QObject *taskObject = sender();
-    QTC_ASSERT(taskObject, return);
-    auto task = static_cast<QFutureWatcher<void> *>(taskObject);
+    const auto it = m_runningTasks.constFind(task);
+    QTC_ASSERT(it != m_runningTasks.constEnd(), return);
+    const Id type = *it;
     if (m_applicationTask == task)
         disconnectApplicationTask();
-    Id type = m_runningTasks.value(task);
-    m_runningTasks.remove(task);
-    delete task;
+    task->disconnect();
+    task->deleteLater();
+    m_runningTasks.erase(it);
     updateSummaryProgressBar();
 
-    if (!m_runningTasks.key(type, 0))
+    if (!m_runningTasks.key(type, nullptr))
         emit allTasksFinished(type);
 }
 
 void ProgressManagerPrivate::disconnectApplicationTask()
 {
+    if (!m_applicationTask)
+        return;
+
     disconnect(m_applicationTask, &QFutureWatcherBase::progressRangeChanged,
                this, &ProgressManagerPrivate::setApplicationProgressRange);
     disconnect(m_applicationTask, &QFutureWatcherBase::progressValueChanged,
@@ -499,7 +796,7 @@ void ProgressManagerPrivate::updateSummaryProgressBar()
     m_summaryProgressBar->setFinished(false);
     static const int TASK_RANGE = 100;
     int value = 0;
-    for (auto it = m_runningTasks.cbegin(), end = m_runningTasks.cend(); it != end; ++it) {
+    for (auto it = m_runningTasks.cbegin(); it != m_runningTasks.cend(); ++it) {
         QFutureWatcher<void> *watcher = it.key();
         int min = watcher->progressMinimum();
         int range = watcher->progressMaximum() - min;
@@ -531,7 +828,7 @@ void ProgressManagerPrivate::stopFadeOfSummaryProgress()
 
 bool ProgressManagerPrivate::hasError() const
 {
-    for (const FutureProgress *progress : qAsConst(m_taskList))
+    for (const FutureProgress *progress : std::as_const(m_taskList))
         if (progress->hasError())
             return true;
     return false;
@@ -541,22 +838,13 @@ bool ProgressManagerPrivate::isLastFading() const
 {
     if (m_taskList.isEmpty())
         return false;
-    for (const FutureProgress *progress : qAsConst(m_taskList)) {
+    for (const FutureProgress *progress : std::as_const(m_taskList)) {
         if (!progress->isFading()) // we still have progress bars that are not fading
             return false;
     }
     return true;
 }
 
-void ProgressManagerPrivate::slotRemoveTask()
-{
-    auto progress = qobject_cast<FutureProgress *>(sender());
-    QTC_ASSERT(progress, return);
-    Id type = progress->type();
-    removeTask(progress);
-    removeOldTasks(type, true);
-}
-
 void ProgressManagerPrivate::removeOldTasks(const Id type, bool keepOne)
 {
     bool firstFound = !keepOne; // start with false if we want to keep one
@@ -592,7 +880,7 @@ void ProgressManagerPrivate::removeOneOldTask()
         Id type = (*i)->type();
 
         int taskCount = 0;
-        for (const FutureProgress *p : qAsConst(m_taskList))
+        for (const FutureProgress *p : std::as_const(m_taskList))
             if (p->type() == type)
                 ++taskCount;
 
@@ -630,6 +918,7 @@ void ProgressManagerPrivate::updateVisibility()
     m_progressView->setVisible(m_progressViewPinned || m_hovered || m_progressView->isHovered());
     m_summaryProgressWidget->setVisible((!m_runningTasks.isEmpty() || !m_taskList.isEmpty())
                                      && !m_progressViewPinned);
+    updateNotificationSummaryIcon();
 }
 
 void ProgressManagerPrivate::updateVisibilityWithDelay()
@@ -652,9 +941,7 @@ void ProgressManagerPrivate::updateStatusDetailsWidget()
         } else if (progress->isSubtitleVisibleInStatusBar() && !progress->subtitle().isEmpty()) {
             if (!m_statusDetailsLabel) {
                 m_statusDetailsLabel = new QLabel(m_summaryProgressWidget);
-                QFont font(m_statusDetailsLabel->font());
-                font.setPointSizeF(StyleHelper::sidebarFontSize());
-                font.setBold(true);
+                const QFont font = StyleHelper::uiFont(StyleHelper::UiElementCaptionStrong);
                 m_statusDetailsLabel->setFont(font);
             }
             m_statusDetailsLabel->setText(progress->subtitle());
@@ -680,6 +967,12 @@ void ProgressManagerPrivate::updateStatusDetailsWidget()
     m_currentStatusDetailsWidget = candidateWidget;
 }
 
+void ProgressManagerPrivate::updateNotificationSummaryIcon()
+{
+    m_notificationSummaryIcon->setVisible(
+        !m_progressViewPinned && !m_infoBarDisplay->infoBar()->entries().isEmpty());
+}
+
 void ProgressManagerPrivate::summaryProgressFinishedFading()
 {
     m_summaryProgressWidget->setVisible(false);
@@ -689,6 +982,8 @@ void ProgressManagerPrivate::summaryProgressFinishedFading()
 void ProgressManagerPrivate::progressDetailsToggled(bool checked)
 {
     m_progressViewPinned = checked;
+    if (!checked)
+        m_hovered = false; // make it take effect immediately even though the mouse is on the button
     updateVisibility();
 
     QtcSettings *settings = ICore::settings();
@@ -697,6 +992,15 @@ void ProgressManagerPrivate::progressDetailsToggled(bool checked)
     settings->endGroup();
 }
 
+void ProgressManagerPrivate::doSetApplicationLabel(const QString &text)
+{
+    if (m_appLabelText == text)
+        return;
+    m_appLabelText = text;
+    if (!m_appLabelUpdateTimer->isActive())
+        m_appLabelUpdateTimer->start(20);
+}
+
 /*!
     \internal
 */
@@ -743,7 +1047,7 @@ FutureProgress *ProgressManager::addTask(const QFuture<void> &future, const QStr
     Shows a progress indicator for task given by the QFutureInterface object
     \a futureInterface.
     The progress indicator shows the specified \a title along with the progress bar.
-    The progress indicator will increase monotonically with time, at \a expectedSeconds
+    The progress indicator will increase monotonically with time, at \a expectedDuration
     it will reach about 80%, and continue to increase with a decreasingly slower rate.
 
     The \a type of a task will specify a logical grouping with other
@@ -753,19 +1057,52 @@ FutureProgress *ProgressManager::addTask(const QFuture<void> &future, const QStr
     \sa addTask
 */
 
-FutureProgress *ProgressManager::addTimedTask(const QFutureInterface<void> &futureInterface, const QString &title,
-                                              Id type, int expectedSeconds, ProgressFlags flags)
+FutureProgress *ProgressManager::addTimedTask(const QFutureInterface<void> &futureInterface,
+                                              const QString &title,
+                                              Id type,
+                                              seconds expectedDuration,
+                                              ProgressFlags flags)
 {
     QFutureInterface<void> dummy(futureInterface); // Need mutable to access .future()
     FutureProgress *fp = m_instance->doAddTask(dummy.future(), title, type, flags);
-    (void) new ProgressTimer(futureInterface, expectedSeconds, fp);
+    (void) new ProgressTimer(futureInterface, expectedDuration, fp);
+    return fp;
+}
+
+FutureProgress *ProgressManager::addTimedTask(const QFuture<void> &future,
+                                              const QString &title,
+                                              Id type,
+                                              seconds expectedDuration,
+                                              ProgressFlags flags)
+{
+    QFutureInterface<void> dummyFutureInterface;
+    QFuture<void> dummyFuture = dummyFutureInterface.future();
+    FutureProgress *fp = m_instance->doAddTask(dummyFuture, title, type, flags);
+    (void) new ProgressTimer(dummyFutureInterface, expectedDuration, fp);
+
+    QFutureWatcher<void> *dummyWatcher = new QFutureWatcher<void>(fp);
+    connect(dummyWatcher, &QFutureWatcher<void>::canceled, dummyWatcher, [future] {
+        QFuture<void> mutableFuture = future;
+        mutableFuture.cancel();
+    });
+    dummyWatcher->setFuture(dummyFuture);
+
+    QFutureWatcher<void> *origWatcher = new QFutureWatcher<void>(fp);
+    connect(origWatcher, &QFutureWatcher<void>::finished, origWatcher, [future, dummyFutureInterface] {
+        QFutureInterface<void> mutableDummyFutureInterface = dummyFutureInterface;
+        if (future.isCanceled())
+            mutableDummyFutureInterface.reportCanceled();
+        mutableDummyFutureInterface.reportFinished();
+    });
+    origWatcher->setFuture(future);
+
     return fp;
 }
 
 /*!
     Shows the given \a text in a platform dependent way in the application
     icon in the system's task bar or dock. This is used to show the number
-    of build errors on Windows 7 and \macos.
+    of build errors on Windows and \macos.
 */
 void ProgressManager::setApplicationLabel(const QString &text)
 {
@@ -784,33 +1121,4 @@ void ProgressManager::cancelTasks(Id type)
         m_instance->doCancelTasks(type);
 }
 
-
-ProgressTimer::ProgressTimer(const QFutureInterfaceBase &futureInterface,
-                             int expectedSeconds,
-                             QObject *parent)
-    : QObject(parent),
-      m_futureInterface(futureInterface),
-      m_expectedTime(expectedSeconds)
-{
-    m_futureInterface.setProgressRange(0, 100);
-    m_futureInterface.setProgressValue(0);
-
-    m_timer = new QTimer(this);
-    m_timer->setInterval(TimerInterval);
-    connect(m_timer, &QTimer::timeout, this, &ProgressTimer::handleTimeout);
-    m_timer->start();
-}
-
-void ProgressTimer::handleTimeout()
-{
-    ++m_currentTime;
-
-    // This maps expectation to atan(1) to Pi/4 ~= 0.78, i.e. snaps
-    // from 78% to 100% when expectations are met at the time the
-    // future finishes. That's not bad for a random choice.
-    const double mapped = atan2(double(m_currentTime) * TimerInterval / 1000.0,
-                                double(m_expectedTime));
-    const double progress = 100 * 2 * mapped / 3.14;
-    m_futureInterface.setProgressValue(int(progress));
-}
-
+} // Core
