@@ -7,6 +7,7 @@
 #include "buildmanager.h"
 #include "deployconfiguration.h"
 #include "kit.h"
+#include "kitaspect.h"
 #include "kitmanager.h"
 #include "project.h"
 #include "projectexplorer.h"
@@ -150,7 +151,7 @@ class GenericModel : public TreeModel<GenericItem, GenericItem>
 public:
     GenericModel(QObject *parent) : TreeModel(parent) { }
 
-    void rebuild(const QList<QObject *> &objects)
+    void rebuild(const QObjectList &objects)
     {
         clear();
         for (QObject * const e : objects)
@@ -249,8 +250,8 @@ public:
     explicit ProjectListView(QWidget *parent = nullptr) : SelectorView(parent)
     {
         const auto model = new GenericModel(this);
-        model->rebuild(transform<QList<QObject *>>(ProjectManager::projects(),
-                                                   [](Project *p) { return p; }));
+        model->rebuild(transform<QObjectList>(ProjectManager::projects(),
+                                              [](Project *p) { return p; }));
         connect(ProjectManager::instance(), &ProjectManager::projectAdded,
                 this, [this, model](Project *project) {
             const GenericItem *projectItem = model->addItemForObject(project);
@@ -322,7 +323,7 @@ signals:
     void changeActiveProjectConfiguration(QObject *pc);
 
 public:
-    void setProjectConfigurations(const QList<QObject *> &list, QObject *active)
+    void setProjectConfigurations(const QObjectList &list, QObject *active)
     {
         theModel()->rebuild(list);
         resetOptimalWidth();
@@ -434,7 +435,7 @@ void TargetSelectorDelegate::paint(QPainter *painter,
     painter->save();
     painter->setClipping(false);
 
-    QColor textColor = creatorTheme()->color(Theme::MiniProjectTargetSelectorTextColor);
+    QColor textColor = creatorColor(Theme::MiniProjectTargetSelectorTextColor);
     if (option.state & QStyle::State_Selected) {
         QColor color;
         if (m_view->hasFocus()) {
@@ -494,7 +495,7 @@ SelectorView::SelectorView(QWidget *parent) : TreeView(parent)
     setSelectionBehavior(SelectRows);
     setAttribute(Qt::WA_MacShowFocusRect, false);
     setHeaderHidden(true);
-    const QColor bgColor = creatorTheme()->color(Theme::MiniProjectTargetSelectorBackgroundColor);
+    const QColor bgColor = creatorColor(Theme::MiniProjectTargetSelectorBackgroundColor);
     const QString bgColorName = creatorTheme()->flag(Theme::FlatToolBars)
             ? bgColor.lighter(120).name() : bgColor.name();
     setStyleSheet(QString::fromLatin1("QAbstractItemView { background: %1; border-style: none; }").arg(bgColorName));
@@ -552,16 +553,15 @@ int SelectorView::padding()
 /////////
 // KitAreaWidget
 /////////
-void doLayout(KitAspect *aspect, Layouting::LayoutItem &builder)
+void doLayout(KitAspect *aspect, Layouting::Layout &builder)
 {
     aspect->addToLayout(builder);
 }
 
 class KitAreaWidget : public QWidget
 {
-    Q_OBJECT
 public:
-    explicit KitAreaWidget(QWidget *parent = nullptr)
+    explicit KitAreaWidget(QWidget *parent)
         : QWidget(parent)
     {
         connect(KitManager::instance(), &KitManager::kitUpdated, this, &KitAreaWidget::updateKit);
@@ -587,7 +587,8 @@ public:
             if (k && k->isMutable(factory->id())) {
                 KitAspect *aspect = factory->createKitAspect(k);
                 m_kitAspects << aspect;
-                grid.addItems({aspect, Layouting::br});
+                grid.addItem(aspect);
+                grid.flush();
             }
         }
         m_gridWidget = grid.emerge();
@@ -599,7 +600,7 @@ public:
     }
 
 private:
-    void updateKit(ProjectExplorer::Kit *k)
+    void updateKit(Kit *k)
     {
         if (!m_kit || m_kit != k)
             return;
@@ -736,33 +737,29 @@ MiniProjectTargetSelector::MiniProjectTargetSelector(QAction *targetSelectorActi
             });
     connect(m_listWidgets[DEPLOY], &GenericListWidget::changeActiveProjectConfiguration,
             this, [this](QObject *pc) {
-                 m_project->activeTarget()->setActiveDeployConfiguration(
+                 m_project->activeBuildConfiguration()->setActiveDeployConfiguration(
                     static_cast<DeployConfiguration *>(pc), SetActive::Cascade);
             });
     connect(m_listWidgets[RUN], &GenericListWidget::changeActiveProjectConfiguration,
             this, [this](QObject *pc) {
-                 m_project->activeTarget()->setActiveRunConfiguration(static_cast<RunConfiguration *>(pc));
+                 m_project->activeBuildConfiguration()->setActiveRunConfiguration(static_cast<RunConfiguration *>(pc));
             });
 }
 
 bool MiniProjectTargetSelector::event(QEvent *event)
 {
-    if (event->type() == QEvent::LayoutRequest) {
-        doLayout(true);
+    if (event->type() == QEvent::ShortcutOverride
+        && static_cast<QKeyEvent *>(event)->key() == Qt::Key_Escape) {
+        event->accept();
         return true;
-    } else if (event->type() == QEvent::ShortcutOverride) {
-        if (static_cast<QKeyEvent *>(event)->key() == Qt::Key_Escape) {
-            event->accept();
-            return true;
-        }
     }
     return QWidget::event(event);
 }
 
 // does some fancy calculations to ensure proper widths for the list widgets
-QVector<int> MiniProjectTargetSelector::listWidgetWidths(int minSize, int maxSize)
+QList<int> MiniProjectTargetSelector::listWidgetWidths(int minSize, int maxSize)
 {
-    QVector<int> result;
+    QList<int> result;
     result.resize(LAST);
     if (m_projectListWidget->isVisibleTo(this))
         result[PROJECT] = m_projectListWidget->optimalWidth();
@@ -801,7 +798,7 @@ QVector<int> MiniProjectTargetSelector::listWidgetWidths(int minSize, int maxSiz
 
     int widthToDistribute = tooSmall ? (minSize - totalWidth)
                                      : (totalWidth - maxSize);
-    QVector<int> indexes;
+    QList<int> indexes;
     indexes.reserve(LAST);
     for (int i = PROJECT; i < LAST; ++i)
         if (result[i] != -1)
@@ -857,7 +854,7 @@ QVector<int> MiniProjectTargetSelector::listWidgetWidths(int minSize, int maxSiz
     }
 }
 
-void MiniProjectTargetSelector::doLayout(bool keepSize)
+void MiniProjectTargetSelector::doLayout()
 {
     // An unconfigured project shows empty build/deploy/run sections
     // if there's a configured project in the seesion
@@ -868,15 +865,15 @@ void MiniProjectTargetSelector::doLayout(bool keepSize)
 
     m_kitAreaWidget->move(0, 0);
 
-    int oldSummaryLabelY = m_summaryLabel->y();
-
-    int kitAreaHeight = m_kitAreaWidget->isVisibleTo(this) ? m_kitAreaWidget->sizeHint().height() : 0;
+    const int kitAreaHeight = m_kitAreaWidget->isVisibleTo(this)
+        ? m_kitAreaWidget->sizeHint().height() : 0;
+    const int kitAreaWidth = m_kitAreaWidget->isVisibleTo(this)
+        ? m_kitAreaWidget->sizeHint().width() : 0;
 
     // 1. Calculate the summary label height
     int summaryLabelY = 1 + kitAreaHeight;
 
     int summaryLabelHeight = 0;
-    int oldSummaryLabelHeight = m_summaryLabel->height();
     bool onlySummary = false;
     // Count the number of lines
     int visibleLineCount = m_projectListWidget->isVisibleTo(this) ? 0 : 1;
@@ -895,9 +892,6 @@ void MiniProjectTargetSelector::doLayout(bool keepSize)
             summaryLabelHeight = m_summaryLabel->sizeHint().height();
     }
 
-    if (keepSize && oldSummaryLabelHeight > summaryLabelHeight)
-        summaryLabelHeight = oldSummaryLabelHeight;
-
     m_summaryLabel->move(0, summaryLabelY);
 
     // Height to be aligned with side bar button
@@ -909,6 +903,7 @@ void MiniProjectTargetSelector::doLayout(bool keepSize)
 
     QRect newGeometry;
 
+    const int minWidth = std::max({m_summaryLabel->sizeHint().width(), kitAreaWidth, 250});
     if (!onlySummary) {
         // list widget height
         int maxItemCount = m_projectListWidget->maxCount();
@@ -916,34 +911,20 @@ void MiniProjectTargetSelector::doLayout(bool keepSize)
             maxItemCount = qMax(maxItemCount, m_listWidgets[i]->maxCount());
 
         int titleWidgetsHeight = m_titleWidgets.first()->height();
-        if (keepSize) {
-            heightWithoutKitArea = height() - oldSummaryLabelY + 1;
-        } else {
-            // Clamp the size of the listwidgets to be at least as high as the sidebar button
-            // and at most half the height of the entire Qt Creator window.
-            heightWithoutKitArea = summaryLabelHeight
-                    + qBound(alignedWithActionHeight,
-                             maxItemCount * 30 + bottomMargin + titleWidgetsHeight,
-                             Core::ICore::mainWindow()->height() / 2);
-        }
+
+        // Clamp the size of the listwidgets to be at least as high as the sidebar button
+        // and at most half the height of the entire Qt Creator window.
+        const int minHeight = alignedWithActionHeight;
+        const int maxHeight = std::max(minHeight, Core::ICore::mainWindow()->height() / 2);
+        heightWithoutKitArea = summaryLabelHeight
+            + qBound(minHeight, maxItemCount * 30 + bottomMargin + titleWidgetsHeight, maxHeight);
 
         int titleY = summaryLabelY + summaryLabelHeight;
         int listY = titleY + titleWidgetsHeight;
         int listHeight = heightWithoutKitArea + kitAreaHeight - bottomMargin - listY + 1;
 
         // list widget widths
-        int minWidth = qMax(m_summaryLabel->sizeHint().width(), 250);
-        minWidth = qMax(minWidth, m_kitAreaWidget->sizeHint().width());
-        if (keepSize) {
-            // Do not make the widget smaller then it was before
-            int oldTotalListWidgetWidth = m_projectListWidget->isVisibleTo(this) ?
-                    m_projectListWidget->width() : 0;
-            for (int i = TARGET; i < LAST; ++i)
-                oldTotalListWidgetWidth += m_listWidgets[i]->width();
-            minWidth = qMax(minWidth, oldTotalListWidgetWidth);
-        }
-
-        QVector<int> widths = listWidgetWidths(minWidth, 1000);
+        QList<int> widths = listWidgetWidths(minWidth, Core::ICore::mainWindow()->width() * 0.9);
 
         const int runColumnWidth = widths[RUN] == -1 ? 0 : RunColumnWidth;
         int x = 0;
@@ -970,13 +951,10 @@ void MiniProjectTargetSelector::doLayout(bool keepSize)
         m_kitAreaWidget->resize(x - 1, kitAreaHeight);
         newGeometry.setSize({x, heightWithoutKitArea + kitAreaHeight});
     } else {
-        if (keepSize)
-            heightWithoutKitArea = height() - oldSummaryLabelY + 1;
-        else
-            heightWithoutKitArea = qMax(summaryLabelHeight + bottomMargin, alignedWithActionHeight);
+        heightWithoutKitArea = qMax(summaryLabelHeight + bottomMargin, alignedWithActionHeight);
         m_summaryLabel->resize(m_summaryLabel->sizeHint().width(), heightWithoutKitArea - bottomMargin);
         m_kitAreaWidget->resize(m_kitAreaWidget->sizeHint());
-        newGeometry.setSize({m_summaryLabel->width() + 1, heightWithoutKitArea + kitAreaHeight});
+        newGeometry.setSize({minWidth + 1, heightWithoutKitArea + kitAreaHeight});
     }
 
     newGeometry.translate(statusBar->mapToGlobal(QPoint{0, 0}));
@@ -1049,10 +1027,6 @@ void MiniProjectTargetSelector::addedTarget(Target *target)
 
     for (BuildConfiguration *bc : target->buildConfigurations())
         addedBuildConfiguration(bc, false);
-    for (DeployConfiguration *dc : target->deployConfigurations())
-        addedDeployConfiguration(dc, false);
-    for (RunConfiguration *rc : target->runConfigurations())
-        addedRunConfiguration(rc, false);
 }
 
 void MiniProjectTargetSelector::removedTarget(Target *target)
@@ -1064,10 +1038,6 @@ void MiniProjectTargetSelector::removedTarget(Target *target)
 
     for (BuildConfiguration *bc : target->buildConfigurations())
         removedBuildConfiguration(bc, false);
-    for (DeployConfiguration *dc : target->deployConfigurations())
-        removedDeployConfiguration(dc, false);
-    for (RunConfiguration *rc : target->runConfigurations())
-        removedRunConfiguration(rc, false);
 }
 
 void MiniProjectTargetSelector::addedBuildConfiguration(BuildConfiguration *bc, bool update)
@@ -1075,6 +1045,11 @@ void MiniProjectTargetSelector::addedBuildConfiguration(BuildConfiguration *bc,
     if (!m_project || bc->target() != m_project->activeTarget())
         return;
 
+    for (DeployConfiguration *dc : bc->deployConfigurations())
+        addedDeployConfiguration(dc, false);
+    for (RunConfiguration *rc : bc->runConfigurations())
+        addedRunConfiguration(rc, false);
+
     m_listWidgets[BUILD]->addProjectConfiguration(bc);
     if (update)
         updateBuildListVisible();
@@ -1085,6 +1060,11 @@ void MiniProjectTargetSelector::removedBuildConfiguration(BuildConfiguration *bc
     if (!m_project || bc->target() != m_project->activeTarget())
         return;
 
+    for (DeployConfiguration *dc : bc->deployConfigurations())
+        removedDeployConfiguration(dc, false);
+    for (RunConfiguration *rc : bc->runConfigurations())
+        removedRunConfiguration(rc, false);
+
     m_listWidgets[BUILD]->removeProjectConfiguration(bc);
     if (update)
         updateBuildListVisible();
@@ -1092,7 +1072,7 @@ void MiniProjectTargetSelector::removedBuildConfiguration(BuildConfiguration *bc
 
 void MiniProjectTargetSelector::addedDeployConfiguration(DeployConfiguration *dc, bool update)
 {
-    if (!m_project || dc->target() != m_project->activeTarget())
+    if (!m_project || dc->buildConfiguration() != m_project->activeBuildConfiguration())
         return;
 
     m_listWidgets[DEPLOY]->addProjectConfiguration(dc);
@@ -1102,7 +1082,7 @@ void MiniProjectTargetSelector::addedDeployConfiguration(DeployConfiguration *dc
 
 void MiniProjectTargetSelector::removedDeployConfiguration(DeployConfiguration *dc, bool update)
 {
-    if (!m_project || dc->target() != m_project->activeTarget())
+    if (!m_project || dc->buildConfiguration() != m_project->activeBuildConfiguration())
         return;
 
     m_listWidgets[DEPLOY]->removeProjectConfiguration(dc);
@@ -1176,8 +1156,10 @@ void MiniProjectTargetSelector::updateDeployListVisible()
     int maxCount = 0;
     for (Project *p : ProjectManager::projects()) {
         const QList<Target *> targets = p->targets();
-        for (Target *t : targets)
-            maxCount = qMax(t->deployConfigurations().size(), maxCount);
+        for (Target *t : targets) {
+            for (const BuildConfiguration * const bc : t->buildConfigurations())
+                maxCount = qMax(bc->deployConfigurations().size(), maxCount);
+        }
     }
 
     bool visible = maxCount > 1;
@@ -1192,8 +1174,10 @@ void MiniProjectTargetSelector::updateRunListVisible()
     int maxCount = 0;
     for (Project *p : ProjectManager::projects()) {
         const QList<Target *> targets = p->targets();
-        for (Target *t : targets)
-            maxCount = qMax(t->runConfigurations().size(), maxCount);
+        for (Target *t : targets) {
+            for (const BuildConfiguration * const bc : t->buildConfigurations())
+                maxCount = qMax(bc->runConfigurations().size(), maxCount);
+        }
     }
 
     bool visible = maxCount > 1;
@@ -1219,13 +1203,13 @@ void MiniProjectTargetSelector::changeStartupProject(Project *project)
     }
 
     if (project) {
-        QList<QObject *> list;
+        QObjectList list;
         const QList<Target *> targets = project->targets();
         for (Target *t : targets)
             list.append(t);
         m_listWidgets[TARGET]->setProjectConfigurations(list, project->activeTarget());
     } else {
-        m_listWidgets[TARGET]->setProjectConfigurations(QList<QObject *>(), nullptr);
+        m_listWidgets[TARGET]->setProjectConfigurations({}, nullptr);
     }
 
     updateActionAndSummary();
@@ -1264,33 +1248,11 @@ void MiniProjectTargetSelector::activeTargetChanged(Target *target)
                    this, &MiniProjectTargetSelector::updateActionAndSummary);
 
     if (m_target) {
-        QList<QObject *> bl;
+        QObjectList bl;
         for (BuildConfiguration *bc : target->buildConfigurations())
             bl.append(bc);
         m_listWidgets[BUILD]->setProjectConfigurations(bl, target->activeBuildConfiguration());
-
-        QList<QObject *> dl;
-        for (DeployConfiguration *dc : target->deployConfigurations())
-            dl.append(dc);
-        m_listWidgets[DEPLOY]->setProjectConfigurations(dl, target->activeDeployConfiguration());
-
-        QList<QObject *> rl;
-        for (RunConfiguration *rc : target->runConfigurations())
-            rl.append(rc);
-        m_listWidgets[RUN]->setProjectConfigurations(rl, target->activeRunConfiguration());
-
-        m_buildConfiguration = m_target->activeBuildConfiguration();
-        if (m_buildConfiguration)
-            connect(m_buildConfiguration, &ProjectConfiguration::displayNameChanged,
-                    this, &MiniProjectTargetSelector::updateActionAndSummary);
-        m_deployConfiguration = m_target->activeDeployConfiguration();
-        if (m_deployConfiguration)
-            connect(m_deployConfiguration, &ProjectConfiguration::displayNameChanged,
-                    this, &MiniProjectTargetSelector::updateActionAndSummary);
-        m_runConfiguration = m_target->activeRunConfiguration();
-        if (m_runConfiguration)
-            connect(m_runConfiguration, &ProjectConfiguration::displayNameChanged,
-                    this, &MiniProjectTargetSelector::updateActionAndSummary);
+        activeBuildConfigurationChanged(target->activeBuildConfiguration());
 
         connect(m_target, &Target::kitChanged,
                 this, &MiniProjectTargetSelector::updateActionAndSummary);
@@ -1303,9 +1265,9 @@ void MiniProjectTargetSelector::activeTargetChanged(Target *target)
         connect(m_target, &Target::activeRunConfigurationChanged,
                 this, &MiniProjectTargetSelector::activeRunConfigurationChanged);
     } else {
-        m_listWidgets[BUILD]->setProjectConfigurations(QList<QObject *>(), nullptr);
-        m_listWidgets[DEPLOY]->setProjectConfigurations(QList<QObject *>(), nullptr);
-        m_listWidgets[RUN]->setProjectConfigurations(QList<QObject *>(), nullptr);
+        m_listWidgets[BUILD]->setProjectConfigurations({}, nullptr);
+        m_listWidgets[DEPLOY]->setProjectConfigurations({}, nullptr);
+        m_listWidgets[RUN]->setProjectConfigurations({}, nullptr);
         m_buildConfiguration = nullptr;
         m_deployConfiguration = nullptr;
         m_runConfiguration = nullptr;
@@ -1321,13 +1283,30 @@ void MiniProjectTargetSelector::kitChanged(Kit *k)
 
 void MiniProjectTargetSelector::activeBuildConfigurationChanged(BuildConfiguration *bc)
 {
-    if (m_buildConfiguration)
+    if (m_buildConfiguration) {
         disconnect(m_buildConfiguration, &ProjectConfiguration::displayNameChanged,
                    this, &MiniProjectTargetSelector::updateActionAndSummary);
+    }
+
     m_buildConfiguration = bc;
     if (m_buildConfiguration)
         connect(m_buildConfiguration, &ProjectConfiguration::displayNameChanged,
                 this, &MiniProjectTargetSelector::updateActionAndSummary);
+    if (m_buildConfiguration) {
+        QObjectList dl;
+        for (DeployConfiguration *dc : bc->deployConfigurations())
+            dl.append(dc);
+        m_listWidgets[DEPLOY]->setProjectConfigurations(dl, bc->activeDeployConfiguration());
+        activeDeployConfigurationChanged(m_buildConfiguration->activeDeployConfiguration());
+        QObjectList rl;
+        for (RunConfiguration *rc : bc->runConfigurations())
+            rl.append(rc);
+        m_listWidgets[RUN]->setProjectConfigurations(rl, bc->activeRunConfiguration());
+        activeRunConfigurationChanged(m_buildConfiguration->activeRunConfiguration());
+    } else {
+        m_listWidgets[DEPLOY]->setProjectConfigurations({}, nullptr);
+        activeDeployConfigurationChanged(nullptr);
+    }
     m_listWidgets[BUILD]->setActiveProjectConfiguration(bc);
     updateActionAndSummary();
 }
@@ -1360,7 +1339,7 @@ void MiniProjectTargetSelector::activeRunConfigurationChanged(RunConfiguration *
 
 void MiniProjectTargetSelector::setVisible(bool visible)
 {
-    doLayout(false);
+    doLayout();
     QWidget::setVisible(visible);
     m_projectAction->setChecked(visible);
     if (visible) {
@@ -1555,7 +1534,7 @@ void MiniProjectTargetSelector::updateSummary()
     }
     if (summary != m_summaryLabel->text()) {
         m_summaryLabel->setText(summary);
-        doLayout(false);
+        doLayout();
     }
 }
 
@@ -1563,7 +1542,7 @@ void MiniProjectTargetSelector::paintEvent(QPaintEvent *)
 {
     QPainter painter(this);
     painter.fillRect(rect(), StyleHelper::baseColor());
-    painter.setPen(creatorTheme()->color(Theme::MiniProjectTargetSelectorBorderColor));
+    painter.setPen(creatorColor(Theme::MiniProjectTargetSelectorBorderColor));
     // draw border on top and right
     QRectF borderRect = QRectF(rect()).adjusted(0.5, 0.5, -0.5, -0.5);
     painter.drawLine(borderRect.topLeft(), borderRect.topRight());
